@@ -1,13 +1,11 @@
 import asyncio
-import click
 import json
-import os
 import sys
+
+import click
 import yaml
 
-import dask.config
 from distributed.deploy.spec import run_spec
-from distributed.utils import deserialize_for_cli
 
 
 @click.command(context_settings=dict(ignore_unknown_options=True))
@@ -17,10 +15,6 @@ from distributed.utils import deserialize_for_cli
 @click.version_option()
 def main(args, spec: str, spec_file: str):
 
-    if "DASK_INTERNAL_INHERIT_CONFIG" in os.environ:
-        config = deserialize_for_cli(os.environ["DASK_INTERNAL_INHERIT_CONFIG"])
-        dask.config.update(dask.config.global_config, config)
-
     if spec and spec_file or not spec and not spec_file:
         print("Must specify exactly one of --spec and --spec-file")
         sys.exit(1)
